<?php
namespace Zeedhi\DevOps\Model\Projectmanagement\Entities;

use Service\TaskPrecursor;
use Zeedhi\DevOps\Util\Utilities;
use Zeedhi\DevOps\Model\Projectmanagement\EntitiesInterface\Log\ProjectLogInterface;
use Zeedhi\DevOps\Model\Projectmanagement\EntitiesInterface\Log\TaskLogInterface;
use Zeedhi\DevOps\Model\Projectmanagement\EntitiesInterface\Tree\Treeble;
use Zeedhi\DevOps\Model\Projectmanagement\EntitiesInterface\Calc\TaskInterface;
use Zeedhi\DevOps\Model\Projectmanagement\EntitiesInterface\Calendar\TaskCalendarInterface;
use Zeedhi\DevOps\Util\Events\Audit_with_id\AuditAble;
use Zeedhi\DevOps\Util\Events\Organization_with_id\OrganizationAble;

class Task extends \Zeedhi\DevOps\Model\Projectmanagement\Entities\Base\Task implements AuditAble, OrganizationAble, Treeble, TaskInterface, TaskCalendarInterface, \Zeedhi\DevOps\Model\Util\LogicalDeleted, ProjectLogInterface, TaskLogInterface
{

    /**
     * Table name to which this class belongs.
     */
    const TABLE = "TASK";
    const ACTIVE = "A";
    const DELETED = 'D';
    const NOT_ACTIVE = "I";
    const IS_PARENT = "Y";
    const NOT_IS_PARENT = "N";
    const IS_ISSUE = "Y";
    const NOT_IS_ISSUE = "N";
    const MEETING_TASK_TYPE = "T";
    const MEETING_SCHEDULE_TYPE = "P";
    const MEETING_ISSUE_TYPE = "I";
    const USED_IN_FRAMING = 'Y';
    const NOT_USED_IN_FRAMING = 'N';
    const IS_NOT_ORIGINAL_BASELINE = 'N';
    const BASELINE_OCCURRENCE_MOVED = 'MOVED';
    const BASELINE_OCCURRENCE_DELETED = 'DELETED';
    const BASELINE_OCCURRENCE_INACTIVATED = 'INACTIVATED';

    /**
     * @var \Service\Calc\TaskInterface[] $children
     */
    private $children;

    /**
     * @var \Closure
     */
    private $childrenClosure;

    /**
     * @var bool
     */
    private $needScheduleCalcDates = true;

    private $dateBeginMax;

    private $dateEndMin;
    /**
     * @var \Service\Calendar\Restriction\Strategy\RestrictionStrategy $restriction
     */
    private $restriction;

    protected $pass = false;

    protected $allCalendars ;

    protected $allDatesPrecursors;

    protected $durationPrecursor = 0;

    public function setParent(\Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $parent = NULL) {
        parent::setParent($parent);
        if($parent){
            $parent->setIsParent('Y');
        }
    }

    public function getTableName(){
        return self::TABLE;
    }

    public function getTreeCode(){
        return $this->getTreeKey();
    }

    public function setTreeCode($key){
        $this->setTreeKey($key);
    }

    public function getTreeParent(){
        return !is_null($this->getParent()) && $this->getParent()->isActive() ? $this->getParent() : $this->getProject() ;
    }

    public function setChildrenClosure(\Closure $childrenClosure){
        $this->childrenClosure = $childrenClosure;
    }

    public function directChildren(){
        return $this->getSubTasks();
    }
    public function children(){
        if ($this->children === null) {
            $this->children = $this->childrenClosure->__invoke();
        }
        return $this->children;
    }
    public function parent(){
        return $this->getTreeParent();
    }
    public function hasParent(){
        return !is_null($this->getTreeParent());
    }
    public function isParent(){
        return $this->hasChildren();
    }
    public function hasChildren(){
        $subTasks = $this->getSubTasks();
        if($subTasks){
            $values = $subTasks->getValues();
            return  !empty($values);
        }else{
            return false;
        }
    }

    public function setPlanned($time){
        $this->setTimePlanned($time);
    }
    public function setSpend($time){
        $this->setTimeSpent($time);
    }
    public function setRemaining($time){
        $this->setTimeRemaining($time);
    }

    public function getPlanned(){
        $timePlanned= $this->getTimePlanned();
        if($this->isTimeLine()){
            $timePlanned = 0;
        }
        return $timePlanned;
    }
    public function getSpend(){
        $timeSpent = $this->getTimeSpent();
        if($this->isTimeLine()){
            $timeSpent = 0;
        }
        return $timeSpent;
    }
    public function getRemaining(){
        $timeRemaining = $this->getTimeRemaining();
        if($this->isTimeLine()){
            $timeRemaining = 0;
        }
        return $timeRemaining;
    }

    public function getEstimatedEndDateMaxParents(){
        if($this->getParent()){
            $return =  $this->getParent()->getEstimatedEndDateMaxParents();
        }else{
            $return = $this->getProject()->getEstimatedEndDateMaxParents();
        }
        return $return;
    }

    public function getEstimatedBeginDateMinParents(){
        if($this->getParent()){
            $return =  $this->getParent()->getEstimatedBeginDateMinParents();
        }else{
            $return = $this->getProject()->getEstimatedBeginDateMinParents();
        }
        return $return;
    }


    public function getAllCalendars()
    {
        if(!$this->allCalendars){
            $this->allCalendars = array_merge($this->getProject()->getAllCalendars(),$this->getParentsCalendar());
        }
        return $this->allCalendars;
    }

    public function getRestriction()
    {
        return $this->restriction;
    }



    public function getParentsCalendar(){
        $calendar = array($this->getCalendar());
        if ($this->getParent()) {
            $calendar = array_merge($this->getParent()->getParentsCalendar(), $calendar);
        }
        $allocations =  $this->getAllocations();
        if (!empty($allocations)) {
            /**
             * @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Allocation[] $allocations
             */
            foreach ($allocations as $allocation) {
                $calendar = array_merge($calendar, $allocation->getAllCalendars());
        }
    }
        return $calendar;
    }

    public function getResourceIds(){
        $allocations =  $this->getAllocations();
        $resources = array();
        if($allocations){
            foreach ($allocations as $allocation) {
                $resources[] = $allocation->getResource()->getId();
            }
        }
        return $resources;
    }

    public function getDatesPrecursors(){
       /** @var \Doctrine\ORM\PersistentCollection $taskRelationships */
        if($this->allDatesPrecursors){
            if($this->getTaskRelationship()){
                $taskRelationships = $this->getTaskRelationship();
                foreach($taskRelationships as $taskRelationship){
                    if($taskRelationship->getType() == TaskPrecursor::PRECURSOR_TYPE){
                        array_push($this->allDatesPrecursors,$taskRelationship->getDatesBeginAndEnd());
                    }
                }
            }
        }else{
            $this->allDatesPrecursors = array();
        }
        return $this->allDatesPrecursors;
    }

    public function getTaskBranchRecursive($branch = null){
        $taskRelationships = $this->getTaskRelationshipInverse();
        $generatedTypeRelationship = \Service\TaskPrecursor::getGeneratedTypRelationship();
        $generatedTypeRelationshipWithoutBranch = \Service\TaskPrecursor::getGeneratedTypRelationshipWithoutBranch();
        if($taskRelationships){
            foreach($taskRelationships as $taskRelationship){
                if(in_array($taskRelationship->getType(), $generatedTypeRelationship) ||
                    $taskRelationship->getType() == TaskRelationship::RELATED_TASK_TYPE) {
                    $mainRelatedTask = \Util\Utilities::getMainRelatedTaskByTaskId($taskRelationship->getTaskRelationship()->getId());
                    $taskToFindBranch = $mainRelatedTask ? $mainRelatedTask : $taskRelationship->getTask();
                    $branches = $taskToFindBranch->getTaskBranches();
                    if($taskRelationship->getType() == TaskRelationship::REPLICATION_TYPE) {
                        $branches = $taskRelationship->getTaskRelationship()->getTaskBranches();
                    }
                    foreach ($branches as $taskBranch) {
                        $branch['BRANCH_NAME'] = $taskBranch->getName();
                        $branch['BRANCH_TASK'] = $taskBranch->getTask()->getId();
                    }
                    if (!$branch) {
                        $branch = $taskRelationship->getTask()->getTaskBranchRecursive();
                    }
                }
            }
            if(!$branch) {
                $branches = $this->getTaskBranches();
                foreach ($branches as $taskBranch) {
                    $branch['BRANCH_NAME'] = $taskBranch->getName();
                    $branch['BRANCH_TASK'] = $taskBranch->getTask()->getId();
                }
            }
        } else {
            $branches = $this->getTaskBranches();
            foreach ($branches as $taskBranch) {
                $branch['BRANCH_NAME'] = $taskBranch->getName();
                $branch['BRANCH_TASK'] = $taskBranch->getTask()->getId();
            }
        }

        return $branch;
    }

    /**
     * @return boolean
     */
    public function isPass()
    {
        return $this->pass;
    }

    /**
     * @param boolean $pass
     */
    public function setPass($pass)
    {
        $this->pass = $pass;
    }

    /**
     * @return boolean
     */
    public function needScheduleCalcDates()
    {
        return $this->needScheduleCalcDates && $this->usesDataControl();
    }

    /**
     * @param boolean $needScheduleCalcDates
     */
    public function setNeedScheduleCalcDates($needScheduleCalcDates)
    {
        $this->needScheduleCalcDates = $needScheduleCalcDates;
    }

    /**
     * @return mixed
     */
    public function getDateBeginMax()
    {
        return $this->dateBeginMax;
    }

    /**
     * @param mixed $dateBeginMax
     */
    public function setDateBeginMax($dateBeginMax)
    {
        $this->dateBeginMax = $dateBeginMax;
    }

    /**
     * @return mixed
     */
    public function getDateEndMin()
    {
        return $this->dateEndMin;
    }

    /**
     * @param mixed $dateEndMin
     */
    public function setDateEndMin($dateEndMin)
    {
        $this->dateEndMin = $dateEndMin;
    }


    public function hasTaskRelationship()
    {
        $taskRelationship = $this->getTaskRelationship();
        return  $taskRelationship != null && !empty($taskRelationship);
    }

    /**
     * @param \Service\Calendar\Restriction\Strategy\RestrictionStrategy $restriction
     * @return mixed|void
     */
    public function setRestriction($restriction)
    {
        $this->restriction = $restriction;
    }

    public function isLogicalDeleted()
    {
        return $this->getStatus() == self::STATUS_LOGICAL_DELETED;
    }

    /**
     * @return int
     */
    public function getDurationPrecursor()
    {
        return $this->getPreviousTime();
    }

    /**
     * @param int $durationPrecursor
     */
    public function setDurationPrecursor($durationPrecursor)
    {
        $this->setPreviousTime($durationPrecursor);
    }

    public function usesDataControl()
    {
        $projectType = $this->project->getProjectType();
        if(empty($projectType)){
            $return = true;
        }else{
            $return = $this->project->getProjectType()->getUsetimeplanning() == \Service\Calendar\TaskCalendar::USES_DATA_CONTROL;
        }
        return $return;
    }

    public function isIssue()
    {
       return $this->getIsIssue() == 'Y';
    }

    public function isContainer()
    {
        return $this->getIsParent() == 'Y';
    }

    public function getTask()
    {
        return $this;
    }

    public function getProjectStatus()
    {
        return $this->getProject()->getProjectStatus();
    }

    /**
     * @param  Task[] $allPrecursors
     * @param bool $first
     * @return Task[]
     * @throws \Exception
     */
    public function getAllPrecursorsRecursive(Array $allPrecursors = array(),$first = true){
        if(!in_array($this,$allPrecursors)) {
            if(!$first){
                array_push($allPrecursors, $this);
            }
            /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskRelationship[] $taskRelationships */
            $taskRelationships = $this->getTaskRelationship();
            foreach ($taskRelationships as $taskRelationship) {
                if ($taskRelationship->getType() == 'P') {
                    $allPrecursors = array_merge($allPrecursors, $taskRelationship->getTaskRelationship()->getAllPrecursorsRecursive($allPrecursors,false));
                }
            }
        }else{
           throw new \Exception('Error :Relacionamento de predeceção circular da tarefa '. $this->getName(). ' para a '
               .array_pop($allPrecursors)->getName());
        }
        return $allPrecursors;
    }

    /**
     * @param  Task[] $allPrecursors
     * @param bool $first
     * @return Task[]
     * @throws \Exception
     */
    public function getAllInversePrecursorsRecursive(Array $allPrecursors = array(),$first = true){
        if(!in_array($this,$allPrecursors)) {
            if(!$first){
                array_push($allPrecursors, $this);
            }
            /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskRelationship[] $taskRelationships */
            $taskRelationships = $this->getTaskRelationshipInverse();
            foreach ($taskRelationships as $taskRelationship) {
                if ($taskRelationship->getType() == 'P') {
                    $allPrecursors = array_merge($allPrecursors, $taskRelationship->getTask()
                                                            ->getAllInversePrecursorsRecursive($allPrecursors,false));
                }
            }
        }else{
           throw new \Exception('Error :Relacionamento de predeceção circular da tarefa '. $this->getName(). ' para a '
               .array_pop($allPrecursors)->getName());
        }
        return $allPrecursors;
    }

    public function getAllParentTasks(array $allParentTasks = array(),$first = true){
        if(!in_array($this,$allParentTasks)) {
            if (!$first) {
                array_push($allParentTasks, $this);
            }
            if ($this->getParent()) {
                $allParentTasks = array_merge($allParentTasks, $this->getParent()->getAllParentTasks($allParentTasks, false));
            }
        }else{
            throw new \Exception('Error :Relacionamento circular da tarefa '. $this->getName(). ' para a '
                .array_pop($allParentTasks)->getName());
        }
        return $allParentTasks;
    }
    public function hasCircularRecursion(){
        try{
            $this->getAllPrecursorsRecursive();
            $return = false;
        }catch (\Exception $e){
            $return = $e->getMessage();
        }
        return $return;
    }

    public function allowEndTask()
    {
        $projectType = $this->getProject()->getProjectType();
        if($projectType){
            $allowEndTask = $projectType->getAllowendtask();
        }else{
            $allowEndTask = 'Y';
        }
        return $allowEndTask == 'Y';
    }

    /**
     * @return Task[]
     */
    public function getTaskInRelationshipInverseTypeIssue(){
        /** @var TaskRelationship[] $taskRelationships */
        $taskRelationships = $this->getTaskRelationshipInverse();
        $tasks = array();
        if($taskRelationships){
            foreach ($taskRelationships as $taskRelationship) {
                if($taskRelationship->getType() == TaskRelationship::RELATIONSHIP_ISSUE_TYPE){
                    $tasks[] = $taskRelationship->getTask();
                }
            }
        }
        return $tasks;
    }
    /**
     * @return Task[]
     */
    public function getTaskInRelationshipTypeIssue(){
        /** @var TaskRelationship[] $taskRelationships */
        $taskRelationships = $this->getTaskRelationship();
        $tasks = array();
        if($taskRelationships){
            foreach ($taskRelationships as $taskRelationship) {
                if($taskRelationship->getType() == TaskRelationship::RELATIONSHIP_ISSUE_TYPE){
                    $tasks[] = $taskRelationship->getTaskRelationship();
                }
            }
        }
        return $tasks;
    }
    public function getTaskForIssue(){
        return $this->getTaskInRelationshipTypeIssue();
    }

    /**
     * @return Task[]
     */
    public function getTaskInRelationshipInverseTypeSchedule(){
        /** @var TaskRelationship[] $taskRelationships */
        $taskRelationships = $this->getTaskRelationshipInverse();
        $tasks = array();
        if($taskRelationships){
            foreach ($taskRelationships as $taskRelationship) {
                if($taskRelationship->getType() == TaskRelationship::SCHEDULE_TYPE){
                    $tasks[] = $taskRelationship->getTask();
                }
            }
        }
        return $tasks;
    }
    /**
     * @return Task[]
     */
    public function getTaskInRelationshipTypeSchedule(){
        /** @var TaskRelationship[] $taskRelationships */
        $taskRelationships = $this->getTaskRelationship();
        $tasks = array();
        if($taskRelationships){
            foreach ($taskRelationships as $taskRelationship) {
                if($taskRelationship->getType() == TaskRelationship::SCHEDULE_TYPE){
                    $tasks[] = $taskRelationship->getTaskRelationship();
                }
            }
        }
        return $tasks;
    }

/**
     * @return Task[]
     */
    public function getTaskInRelationshipInverseTypeMergeRequest(){
        /** @var TaskRelationship[] $taskRelationships */
        $taskRelationships = $this->getTaskRelationshipInverse();
        $tasks = array();
        if($taskRelationships){
            foreach ($taskRelationships as $taskRelationship) {
                if($taskRelationship->getType() == TaskRelationship::MERGE_REQUEST_TYPE){
                    $tasks[] = $taskRelationship->getTask();
                }
            }
        }
        return $tasks;
    }
    /**
     * @return Task[]
     */
    public function getTaskInRelationshipTypeMergeRequest(){
        /** @var TaskRelationship[] $taskRelationships */
        $taskRelationships = $this->getTaskRelationship();
        $tasks = array();
        if($taskRelationships){
            foreach ($taskRelationships as $taskRelationship) {
                if($taskRelationship->getType() == TaskRelationship::MERGE_REQUEST_TYPE){
                    $tasks[] = $taskRelationship->getTaskRelationship();
                }
            }
        }
        return $tasks;
    }


    /**
     * @return bool
     */
    public function validateIsNotAllowEndTask()
    {
        return !$this->isIssue()
        && $this->getProject()
        && $this->getProject()->getProjectType()
        && $this->getProject()->getProjectType()->getAllowendtask() == ProjectType::IS_NOT_ALLOW_END_TASK
        && !($this->getClassification()
            && $this->getClassification()->getControl() == Classification::CONTROL_CLIENT_ACCEPT);
    }



    public function setPercentCompleted($percentCompleted)
    {
        if ($this->validateIsNotAllowEndTask()) {
            $percentCompleted = 1;
        }
        $this->percentCompleted = $percentCompleted;
    }

    public function setTaskStatus(TaskStatus $taskStatus = NULL) {
        if ($taskStatus
            && $this->validateIsNotAllowEndTask()
            && ($taskStatus->getControl() == TaskStatus::TASK_STATUS_CONTROL_CONCLUDED
            || $taskStatus->getUsedInTask() == TaskStatus::USED_IN_TASK_NO)
        ) {
            $taskStatus = Utilities::getTaskStatusPaused();
        }
        $this->taskStatus = $taskStatus;
    }

    public function isTask()
    {
        return true;
    }

    public function isActive() {
        return $this->getStatus() == self::ACTIVE && !$this->isBackLog();
    }
    public function isBackLog() {
        return $this->getTaskStatus() && $this->getTaskStatus()->getControl() == TaskStatus::TASK_STATUS_CONTROL_BACKLOG ;
    }

    public function isSchedule()
    {
        $project =  $this->getProject();
        return $project->isMeeting() || $project->isMeetingOccurrences();
    }

    /**
     * @param TaskInterface $child
     * @return bool
     */
    public function isValidateOfCalc($child)
    {
        return  ($this->getProject()->getId() == $child->getProject()->getId() && $child->getIsParent() != 'Y');
    }
    public function isTimeLine()
    {
        return $this->getProject() &&
        $this->getProject()->getProjectType() &&
        $this->getProject()->getProjectType()->getExternaltype() == ProjectType::TYPE_PROJECT_CONTROL_TIME_LINE;
    }


    public function deleteEntity()
    {
        $this->setStatus(self::STATUS_LOGICAL_DELETED);
    }


}
